require_relative '../wendelin_client'
require_relative '../filelogger'

module Embulk
  module Input

    class Fifinput < InputPlugin

      TASK_REPORT_FILE = ".ingestion-task-report"
      COMPLETED_FILE = ".ingestion-completed"

      RUN_DONE = "done"
      RUN_ERROR = "error"
      RUN_ABORTED = "aborted"

      Plugin.register_input("fif", self)

      def self.getTaskReportFilename(data_set_directory)
	return data_set_directory + TASK_REPORT_FILE
      end

      def self.getCompletedFilename(data_set_directory)
	return data_set_directory + COMPLETED_FILE
      end

      EOF = "EOF"
      CHUNK_SIZE = 50000000 #50mb
      MEGA = 1000000
      SCHEMA = [
		 {"name"=>"supplier", "type"=>"string"},
		 {"name"=>"data_set", "type"=>"string"},
		 {"name"=>"file", "type"=>"string"},
		 {"name"=>"extension", "type"=>"string"},
		 {"name"=>"data_chunk", "type"=>"string"},
		 {"name"=>"eof", "type"=>"string"}
               ]

      def self.filterDoneTasks(files, task_report_file, data_set, root_path)
	data_set = data_set.end_with?("/") ? data_set : data_set + "/"
	ingested_files = []
	File.readlines(task_report_file).each do |line|
	  record = line.split(";")
	  if record[1].chomp == RUN_DONE
	    ingested = record[0].sub(data_set, root_path)
	    ingested_files.push(ingested)
	  end
	end
	pending_files = []
	files.each do |file|
	  if ! ingested_files.include? file
	    pending_files.push(file)
	  end
	end
	return pending_files
      end

      def self.filterExistingFiles(files, data_streams, data_set, root_path)
	data_set = data_set.end_with?("/") ? data_set : data_set + "/"
	existing_files = []
	data_streams["result"].each do |data_stream|
	  file_path = root_path + data_stream[1].reverse.sub("/".reverse, ".".reverse).reverse.sub(data_set, "")
	  if file_path.end_with?(".none")
	    file_path = file_path[0...-5]
	  end
	  existing_files.push(file_path)
	end
	filtered_files = files - existing_files
	ignored_files = files - filtered_files
	if not ignored_files.empty?
	  puts
	  @logger.info("There are files in the local directory that already exist in data set and will be ignored for ingestion.", print=TRUE)
	  puts
	end
	if filtered_files.empty?
	  @logger.info("The dataset directory '#{root_path}' does not contain any new file for the data set '#{data_set}'.", print=TRUE)
	  @logger.info("Please make sure your dataset directory contains new files for ingestion.", print=TRUE)
	  @logger.abortExecution()
	end
	return filtered_files
      end

      def self.transaction(config, &control)
	begin
	  tool_dir = config.param('tool_dir', :string)
	  @logger = LogManager.instance()
	  @logger.setFilename(tool_dir, "ingestion")
	  task = { 'paths' => [] }
	  task['supplier'] = config.param('supplier', :string)
	  task['data_set'] = config.param('data_set', :string)
	  task['chunk_size'] = config.param('chunk_size', :float, default: 0) * MEGA
	  if task['chunk_size'] == 0
	    task['chunk_size'] = CHUNK_SIZE
	  end
	  paths = config.param('path_prefix', :array)
	  paths[0] = paths[0].end_with?("/") ? paths[0] : paths[0] + "/"
	  task['inputs'] = paths
	  task['paths'] = paths.map {|path|
				      next [] unless Dir.exist?(path)
				      Dir[(path + '/**/*').gsub! '//', '/']
				    }.flatten.select{ |file| File.file?(file) }

	  @wendelin = WendelinClient.new(config.param('erp5_url', :string), config.param('user', :string), config.param('password', :string))
	  data_stream_list = @wendelin.getDataStreams(task['data_set'])
	  if data_stream_list["status_code"] == 0
	    task['data_streams'] = data_stream_list["result"]
	  else
	    @logger.error(data_stream_list["error_message"], print=TRUE)
	    @logger.abortExecution()
          end

	  @logger.info("Supplier: #{task['supplier']}")
	  @logger.info("Dataset name: #{task['data_set']}")
	  @logger.info("Chunk size set in #{task['chunk_size']/MEGA}MB")
	  if task['data_set'] == '$DATA_SET'
	    @logger.error("There was an error setting the configuration file", print=TRUE)
            @logger.info("Please try manual ingestion or update manually the ingestion configuration file.", print=TRUE)
	    @logger.abortExecution()
	  end
	  if task['paths'].empty?
	    @logger.error("The dataset directory '#{task['inputs'][0]}' is empty.", print=TRUE)
	    @logger.error("Could not find any valid file.", print=TRUE)
	    @logger.error("Please make sure your dataset directory contains files for ingestion.", print=TRUE)
	    @logger.abortExecution()
          end

	  @data_set_directory = paths[0]
	  task_report_file = getTaskReportFilename(@data_set_directory)
	  completed_file = getCompletedFilename(@data_set_directory)
	  delete_task_report = FALSE
	  if File.file?(task_report_file)
	    task['paths'] = filterDoneTasks(task['paths'], task_report_file, task['data_set'], @data_set_directory)
	    if File.file?(completed_file)
	      delete_task_report = TRUE
	    else
	      puts
	      @logger.info("There was a previous attempt to ingest this dataset but it did not finish successfully.", print=TRUE)
	      @logger.info("Resuming ingestion...", print=TRUE)
	    end
	  else
	    File.open(task_report_file, 'w') {}
	  end

	  task['paths'] = filterExistingFiles(task['paths'], data_stream_list, task['data_set'], @data_set_directory)
	  @logger.info("#{task['paths'].length} new file(s) detected for ingestion: ", print=TRUE)
	  if task['paths'].length > 15
	    @logger.info(task['paths'][0, 5], print=TRUE)
	    @logger.info(".....", print=TRUE)
	    @logger.info(task['paths'][task['paths'].length-5, task['paths'].length-1], print=TRUE)
	  else
	    @logger.info(task['paths'], print=TRUE)
	  end
	  puts
	  @logger.info("Continue with ingestion? (y/n)", print=TRUE)
          option = gets
          option = option.chomp
	  if option == "n"
	    @logger.info("Ingestion cancelled by user.", print=TRUE)
	    @logger.abortExecution()
	  end

	  if delete_task_report
	    File.delete(task_report_file) if File.exist?(task_report_file)
	    File.open(task_report_file, 'w') {}
	  end
	  File.delete(completed_file) if File.exist?(completed_file)

	  columns = [
            Column.new(0, "supplier", :string),
            Column.new(1, "data_set", :string),
	    Column.new(2, "file", :string),
	    Column.new(3, "extension", :string),
	    Column.new(4, "data_chunk", :string),
	    Column.new(5, "eof", :string)
          ]

	  commit_reports = yield(task, columns, task['paths'].length)
          done = commit_reports.map{|hash| hash["done"]}.flatten.compact

          resume(task, columns, task['paths'].length, &control)
	rescue Exception => e
	  @logger.error("An error occurred during operation: " + e.to_s, print=TRUE)
	  @logger.error(e.backtrace)
	  puts "[INFO] For more detailed information, please refer to the log file."
	  @logger.abortExecution()
	end
      end

      def self.resume(task, columns, count, &control)
	@logger = LogManager.instance()
        task_reports = yield(task, columns, count)
	@logger.info("Reports:", print=TRUE)
	if task_reports.length > 15
	  @logger.info(task_reports[0, 5], print=TRUE)
	  @logger.info(".....", print=TRUE)
	  @logger.info(task_reports[task_reports.length-5, task_reports.length-1], print=TRUE)
	  @logger.info("Full task report:")
	  @logger.info(task_reports)
	else
	  @logger.info(task_reports, print=TRUE)
	end
        next_config_diff = task_reports.map{|hash| hash["done"]}.flatten.compact
	@logger.info("#{next_config_diff.length} file(s) ingested.", print=TRUE)
	#if next_config_diff.length > 0
	#  @logger.info("FIF INPUT - Your ingested files will be available in the site in a few minutes. Thank for your patience.", print=TRUE)
	#end
	if(next_config_diff.length == count)
	  @logger.info("Dataset successfully ingested.", print=TRUE)
	  completed_file = getCompletedFilename(@data_set_directory)
	  File.open(completed_file, 'w') {}
	else
	  next_config_diff = task_reports.map{|hash| hash["error"]}.flatten.compact
	  puts
	  @logger.error("The following files could not be ingested. Please check the details in the log file: " + @logger.getLogPath(), print=TRUE)
	  if next_config_diff.length > 15
	    @logger.error(next_config_diff[0, 5], print=TRUE)
	    @logger.error(".....", print=TRUE)
	    @logger.error(next_config_diff[next_config_diff.length-5, next_config_diff.length-1], print=TRUE)
	  else
	    @logger.error(next_config_diff, print=TRUE)
	  end
	  @logger.info("You can retry the ingestion for those pending files.", print=TRUE)
	  puts
	end
	next_config_diff = {}
	return {"done" => next_config_diff}        
      end

      def initialize(task, schema, index, page_builder)
        super
        @supplier = task['supplier']
        @dataset = task['data_set']
	@chunk_size = task['chunk_size']
	@input_dirs = task['inputs']
	@data_set_directory = task['inputs'][0]
	@logger = LogManager.instance()
      end

      def run
	path = task['paths'][@index]
	@logger.info("Processing file #{path.to_s}", print=TRUE)
	begin
	  each_chunk(path, schema[1..-1].map{|elm| elm.name}, @chunk_size) do |entry|
	    @page_builder.add(entry)
          end
          @page_builder.finish
	rescue java.lang.OutOfMemoryError
	  @logger.logOutOfMemoryError(path)
	  return_value = RUN_ABORTED
	rescue Exception => e
	  @logger.error("An error occurred during file ingestion: " + e.to_s, print=TRUE)
	  @logger.error(e.backtrace)
	  puts "[INFO] For more detailed information, please refer to the log file: " + @logger.getLogPath()
	  return_value = RUN_ERROR
	else
	  return_value = RUN_DONE
	end
	task_report_file = Fifinput.getTaskReportFilename(@data_set_directory)
	dataset_file = path.sub(@data_set_directory, @dataset.end_with?("/") ? @dataset : @dataset + "/")
	File.open(task_report_file, 'ab') { |file| file.puts(dataset_file+";"+return_value+";") }
	return {return_value => path}
      end

      private

      def each_chunk(path, fields, chunk_size=CHUNK_SIZE)
	extension = File.extname path

	if path.start_with?(@input_dirs[0])
	  filename = path.sub(@input_dirs[0], "")
	  filename = filename.reverse.sub(extension.reverse, "").reverse
	end

	extension.gsub! '.', ''
	if extension == ""
	  extension = "none"
	end
        file_object = File.open(path, "rb")
	npart = 0
        next_byte = file_object.read(1)
	first = TRUE
        while true
            data = next_byte
            if not next_byte
		if first
		  values = [@supplier, @dataset, filename, extension, "", EOF]
		  yield(values)
		end
		break
	    end
	    first = FALSE
            data += file_object.read(chunk_size)
            next_byte = file_object.read(1)
            if not next_byte
                eof = EOF
	    else
		npart += 1
       	 	eof = npart.to_s.rjust(3, "0")
	    end
	    content = Base64.encode64(data)
	    values = [@supplier, @dataset, filename, extension, content, eof]
	    yield(values)
	end
      end
    end
  end
end
